defmodule Mix.Tasks.Mfc.Attachments do
  use Mix.Task
  import Mix.Pleroma
  alias Pleroma.Object

  @moduledoc """
  Module works with object attachments.

  ## Adds source key for attachments which don't have it.

  ## Add source key for image attachments.

      mix mfc.attachments images [OPTION...]

  ## Add source key for video attachments.

      mix mfc.attachments videos [OPTION...]

  ## Add source key for gifs attachments.

      mix mfc.attachments gifs [OPTION...]

  Options:
  - `--limit LIMIT` - limit objects for reencoding  (i.e., 5)

  """

  def run(["images" | args]) do
    start_pleroma()
    process_objects("image/jpeg", args)
  end

  def run(["videos" | args]) do
    start_pleroma()
    process_objects("video/mp4", args)
  end

  def run(["gifs" | args]) do
    start_pleroma()
    process_objects("image/gif", args)
  end

  defp process_objects(type, args) when is_binary(type) do
    {opts, _} = OptionParser.parse!(args, strict: [limit: :integer])

    Object.load_attachments(type, opts)
    |> process_objects(type)
  end

  defp process_objects([], _), do: nil

  defp process_objects([object | others], type) do
    original_path = get_original_path(object, type)

    meta = update_meta(object, original_path <> "_original")

    update_object(object, meta)

    process_objects(others, type)
  end

  defp get_original_path(object, "image/jpeg") do
    object.data["meta"]["original"]["dest_key"]
  end

  defp get_original_path(object, type) when type in ["video/mp4", "image/gif"] do
    url = hd(object.data["url"])
    uri = URI.parse(url["href"])
    [original_path | _] = Enum.reverse(Path.split(uri.path))
    original_path
  end

  defp update_meta(object, original_path) do
    Map.get(object.data, "meta", %{})
    |> Map.put_new("source_key", original_path)
  end

  defp update_object(object, meta) do
    data = Map.put(object.data, "meta", meta)

    Object.update_data(object, data)
  end
end
