defmodule Mix.Tasks.Mfc.Reencode do
  use Mix.Task
  import Mix.Pleroma
  alias Pleroma.Object
  alias Pleroma.Uploaders.MFC

  @shortdoc "Reencode MFC attachments"
  @moduledoc """
  Reencode MFC attachments.

  ## Reencode.

      mix mfc.reencode images [OPTION...]
      mix mfc.reencode videos [OPTION...]

  Options:
  - `--limit LIMIT` - limit objects for reencoding  (i.e., 5)
  """

  def run(["images" | args]) do
    start_pleroma()
    process_objects("image/jpeg", args)
  end

  def run(["videos" | args]) do
    start_pleroma()
    process_objects("video/mp4", args)
  end

  def run(["gifs" | args]) do
    start_pleroma()
    process_objects("image/gif", args)
  end

  defp process_objects(type, args) when is_binary(type) do
    {opts, _} = OptionParser.parse!(args, strict: [limit: :integer])

    shell_info("Starting reencoding #{type} attachments")

    objects = Object.load_attachments(type, opts)

    shell_info("Found #{length(objects)} objects with attachments")
    process_objects(objects, type)
  end

  defp process_objects([], _), do: nil

  defp process_objects([object | others], type) do
    shell_info("Processing #{object.id} object")

    meta = process_object(object, type)

    if meta, do: update_object(object, meta)

    process_objects(others, type)
  end

  defp process_object(object, "image/jpeg") do
    original = object.data["meta"]["source_key"]

    with {:ok, versions} <-
           MFC.Image.convert(
             MFC.Image.Client.client(),
             String.trim_trailing(original, "_original")
           ) do
      shell_info("Processed #{original}")

      Map.get(versions, "versions", %{})
      |> Map.put("source_key", original)
    else
      _ -> nil
    end
  end

  defp process_object(object, type) when type in ["video/mp4", "image/gif"] do
    original = object.data["meta"]["source_key"]

    with path <- original,
         {:ok, _} = Registry.register(MFC, path, path),
         :ok <- MFC.Video.convert(MFC.Video.Client.client(), path),
         {:ok, conversion_result} <- MFC.wait_for_conversion() do
      shell_info("Processed #{original}")

      Map.get(conversion_result, :meta, %{})
      |> Map.put("source_key", original)
    else
      _ -> nil
    end
  end

  defp update_object(object, meta) do
    data = Map.put(object.data, "meta", meta)

    Object.update_data(object, data)
  end
end
