defmodule Mix.Tasks.Mfc.ReplaceAttachmentUrlPrefix do
  use Mix.Task
  import Mix.Pleroma
  import Ecto.Query

  alias Pleroma.Object
  alias Pleroma.Repo

  @shortdoc "Replace url prefix in attachments"

  @moduledoc """
  Replace url prefix in attachments

  ## Usage

      mix mfc.replace_attachment_url_prefix CURRENT_URL_PREFIX NEW_URL_PREFIX

  Example: `mix mfc.replace_attachment_url_prefix "https://example.com" "https://cdn.example.com"`
  """

  def run([from, to]) do
    show_warnings()
    shell_info("Search `#{from}` and replace to `#{to}` in attachment urls.")

    if shell_prompt("Continue?", "n") in ~w(Yn Y y) do
      start_pleroma()
      process(from, to)
      shell_info("Done")
    end
  end

  def run([from, to, ap_id]) do
    show_warnings()

    shell_info("Selected activity: #{ap_id}")
    shell_info("Search `#{from}` and replace to `#{to}` in attachment urls.")

    if shell_prompt("Continue?", "n") in ~w(Yn Y y) do
      start_pleroma()
      process(from, to, ap_id)
      shell_info("Done")
    end
  end

  defp show_warnings do
    [:red, :bright, "==> WARNING! This is A VERY DANGEROUS operation!"]
    |> IO.ANSI.format()
    |> IO.puts()

    [:yellow, "==> Before proceeding make sure you have a database backup"]
    |> IO.ANSI.format()
    |> IO.puts()
  end

  def process(from, to) do
    stream =
      Object
      |> where(fragment("data ->> 'type' = 'Document' AND data ->> 'url' IS NOT NULL"))
      |> or_where(fragment("data ->> 'type' = 'Note' AND data ->> 'attachment' != '[]'"))
      |> Repo.stream()

    Repo.transaction(fn ->
      Enum.each(stream, &update_object(&1, from, to))
    end)
  end

  def process(from, to, activity_id) do
    activity = Pleroma.Activity.get_by_id_with_object(activity_id)
    update_object(activity.object, from, to)
  end

  defp update_object(%{data: %{"url" => _}} = object, from, to) do
    object
    |> Object.change(%{data: update_url(object.data, from, to)})
    |> Repo.update()
  end

  defp update_object(%{data: %{"attachment" => _}} = object, from, to) do
    hrefs =
      Enum.flat_map(object.data["attachment"], fn %{"url" => url} ->
        Enum.map(url, & &1["href"])
      end)

    attachment = Enum.map(object.data["attachment"], &update_url(&1, from, to))
    content = update_content(object.data["content"], hrefs, from, to)

    data =
      object.data
      |> Map.put("attachment", attachment)
      |> Map.put("content", content)

    object
    |> Object.change(%{data: data})
    |> Repo.update()
  end

  defp update_url(%{"type" => "Document"} = data, from, to) do
    url =
      Enum.map(
        data["url"],
        &Map.put(&1, "href", String.replace_prefix(&1["href"], from, to))
      )

    Map.put(data, "url", url)
  end

  defp update_url(data, _from, _to), do: data

  def update_content(content, hrefs, from, to) do
    Enum.reduce(hrefs, content, fn href, content ->
      new_href = String.replace_prefix(href, from, to)
      String.replace(content, href, new_href)
    end)
  end
end
