# Pleroma: A lightweight social networking server
# Copyright © 2019 Pleroma Authors <https://pleroma.social/>
# SPDX-License-Identifier: AGPL-3.0-only

defmodule Pleroma.Web.ActivityPub.MRF.CdnWarmingPolicy do
  @moduledoc "Triggers the CDN Master to warm the global CDN edge nodes"
  @behaviour Pleroma.Web.ActivityPub.MRF

  alias Pleroma.HTTP
  alias Pleroma.Workers.BackgroundWorker

  require Logger

  @hackney_options [
    pool: :media,
    recv_timeout: 10_000
  ]

  def perform(:prefetch, urls) do
    config = get_config()
    log_level = get_log_level(config)

    case get_cdn_trigger_endpoint(config) do
      nil ->
        Logger.log(
          :error,
          "config :pleroma, Pleroma.Web.ActivityPub.MRF.CdnWarmingPolicy, :cdn_trigger_endpoint is not set"
        )

      {endpoint_host, endpoint_path, endpoint_port, endpoint_scheme} ->
        Logger.log(log_level, "Prefetching #{inspect(urls)}")

        Enum.each(urls, fn
          url when is_binary(url) ->
            %URI{host: asset_host, path: asset_path} = URI.parse(url)

            warming_url =
              "#{endpoint_scheme}://#{endpoint_host}:#{endpoint_port}#{endpoint_path}#{asset_path}"

            HTTP.get(warming_url, [{"HOST", asset_host}], adapter: @hackney_options)

          _ ->
            :ok
        end)
    end
  end

  def perform(:preload, %{"object" => %{"attachment" => attachments}} = _message) do
    Enum.each(attachments, fn
      %{"url" => url} when is_list(url) ->
        url
        |> Enum.each(fn
          %{"href" => _href} = attrs ->
            BackgroundWorker.enqueue("cdn_prefetch", %{"url" => get_urls(attrs)})

          x ->
            Logger.debug("Unhandled attachment URL object #{inspect(x)}")
        end)

      x ->
        Logger.debug("Unhandled attachment #{inspect(x)}")
    end)
  end

  @spec get_urls(map()) :: list(String.t())
  defp get_urls(%{"href" => href} = attrs) do
    ([href] ++ [get_preview_url(attrs)])
    |> Enum.uniq()
    |> Enum.filter(& &1)
  end

  @spec get_preview_url(map()) :: String.t()
  defp get_preview_url(%{"href" => href} = attrs) do
    case attrs["preview_url"] do
      nil ->
        attachment_type = attrs["mediaType"] || attrs["mimeType"] || "image"
        Pleroma.Uploaders.Uploader.preview_url(attachment_type, href)

      url ->
        url
    end
  end

  defp get_log_level(config) do
    case Keyword.get(config, :log_level, :info) do
      level when level in [:info, :debug, :warn, :error] -> level
      _ -> :info
    end
  end

  defp get_config, do: Pleroma.Config.get(__MODULE__)

  defp get_cdn_trigger_endpoint(config) do
    %URI{host: endpoint_host, path: endpoint_path, port: endpoint_port, scheme: endpoint_scheme} =
      URI.parse(Keyword.get(config, :cdn_trigger_endpoint, nil))

    case endpoint_host do
      endpoint_host
      when is_nil(endpoint_host) or endpoint_host == "" ->
        nil

      endpoint_host
      when is_binary(endpoint_host) ->
        {endpoint_host, endpoint_path, endpoint_port, endpoint_scheme}

      _ ->
        nil
    end
  end

  @impl true
  def filter(
        %{"type" => "Create", "object" => %{"attachment" => attachments} = _object} = message
      )
      when is_list(attachments) and length(attachments) > 0 do
    BackgroundWorker.enqueue("cdn_preload", %{"message" => message})

    {:ok, message}
  end

  @impl true
  def filter(message), do: {:ok, message}

  @impl true
  def describe, do: {:ok, %{}}

  @impl true
  def config_description do
    %{
      key: :mrf_cdn_warming,
      related_policy: "Pleroma.Web.ActivityPub.MRF.CdnWarmingPolicy",
      label: "MRF CDN Warming Policy",
      description: "Settings for CDN Warming Policy",
      children: [
        %{
          key: :cdn_trigger_endpoint,
          type: :string,
          description: "CDN trigger endpoint",
          suggestions: ["https://cdn-example.com"]
        },
        %{
          key: :log_level,
          type: :atom,
          description: "Log level",
          suggestions: [:debug, :info, :error, :warn]
        }
      ]
    }
  end
end
