defmodule Mix.Tasks.Mfc.AttachmentsTest do
  use Pleroma.DataCase, async: true
  import Pleroma.Factory
  alias Mix.Tasks.Mfc.Attachments
  alias Pleroma.Object

  setup do
    user = insert(:user)

    image_meta = %{
      "small" => %{
        "size" => "333x400",
        "width" => 333,
        "aspect" => 0.8325,
        "height" => 400,
        "method" => "resize",
        "dest_key" =>
          "0619b7f00f1e46653bbb6b79f786294696bcdccc7b30f7f4028bbeca7c2e4540.jpg.preview.jpg",
        "content_type" => "image/jpeg"
      },
      "original" => %{
        "size" => "720x866",
        "width" => 720,
        "aspect" => 0.8314087759815243,
        "height" => 866,
        "method" => "resize",
        "dest_key" => "0619b7f00f1e46653bbb6b79f786294696bcdccc7b30f7f4028bbeca7c2e4540.jpg",
        "content_type" => "image/jpeg"
      }
    }

    object_data = %{
      "url" => [
        %{
          "href" =>
            "https://social.dome.com/media/0619b7f00f1e46653bbb6b79f786294696bcdccc7b30f7f4028bbeca7c2e4540.jpg",
          "type" => "Link",
          "mediaType" => "image/jpeg"
        }
      ],
      "meta" => image_meta,
      "name" => "ZDv317OsNojizA.jpg",
      "type" => "Document",
      "actor" => user.ap_id
    }

    video_meta = %{
      "fps" => 25,
      "size" => "720x408",
      "small" => %{
        "size" => "720x408",
        "width" => 720,
        "aspect" => 1.7647058823529411,
        "height" => 408
      },
      "width" => 720,
      "aspect" => 1.7647058823529411,
      "height" => 408,
      "duration" => 33.53,
      "original" => %{
        "width" => 720,
        "height" => 408,
        "bitrate" => 543_297,
        "duration" => 33.53,
        "frame_rate" => "25/1"
      },
      "audio_encode" => "aac (mp4a / 0x6134706d)",
      "audio_bitrate" => "44100 Hz",
      "audio_channels" => "stereo"
    }

    video_data = %{
      "url" => [
        %{
          "href" =>
            "https://social.dome.com/media/c67a114ceff7817f03dae55fc49c63e691ea38dfd36ba9f761877186cba47f12.mp4",
          "type" => "Link",
          "mediaType" => "video/mp4"
        }
      ],
      "meta" => video_meta,
      "name" => "BMbW85Vze3DYg.mp4",
      "type" => "Document",
      "actor" => user.ap_id
    }

    new_gif_data = %{
      "url" => [
        %{
          "href" =>
            "https://social.dome.com/media/d7c016123f31b86936a2046c42cc8380036fbcd6d008ba0c2801eb73884ce8b9.mp4",
          "type" => "Link",
          "mediaType" => "image/gif"
        }
      ],
      "meta" => %{},
      "name" => "QgBJ-6DXX5xRMw.gif",
      "type" => "Document",
      "actor" => user.ap_id
    }

    old_gif_data = %{
      "url" => [
        %{
          "href" =>
            "https://social.myfreecams.com/media/dd02ebe7ebfa42777f3265794c2274ad85447d40731908a78f6288f7dcdd930b.gif",
          "type" => "Link",
          "mediaType" => "image/gif"
        }
      ],
      "name" => "QgBJ-6DXX5xRMw.gif",
      "type" => "Document",
      "actor" => user.ap_id
    }

    {:ok,
     image_meta: image_meta,
     object_data: object_data,
     video_meta: video_meta,
     video_data: video_data,
     new_gif_data: new_gif_data,
     old_gif_data: old_gif_data}
  end

  describe "adding source key" do
    test "to image", context do
      {:ok, object} = Repo.insert(%Object{data: context.object_data})
      Attachments.run(["images"])

      reloaded = Repo.get(Object, object.id)

      assert reloaded.data["meta"] ==
               Map.put(
                 context.image_meta,
                 "source_key",
                 "0619b7f00f1e46653bbb6b79f786294696bcdccc7b30f7f4028bbeca7c2e4540.jpg_original"
               )
    end

    test "to video", context do
      {:ok, object} = Repo.insert(%Object{data: context.video_data})
      Attachments.run(["videos"])

      reloaded = Repo.get(Object, object.id)

      assert reloaded.data["meta"] ==
               Map.put(
                 context.video_meta,
                 "source_key",
                 "c67a114ceff7817f03dae55fc49c63e691ea38dfd36ba9f761877186cba47f12.mp4_original"
               )
    end

    test "to gifs", context do
      {:ok, new_gif_object} = Repo.insert(%Object{data: context.new_gif_data})
      {:ok, old_gif_object} = Repo.insert(%Object{data: context.old_gif_data})
      Attachments.run(["gifs"])

      reloaded_new = Repo.get(Object, new_gif_object.id)
      reloaded_old = Repo.get(Object, old_gif_object.id)

      assert reloaded_new.data["meta"] ==
               %{
                 "source_key" =>
                   "d7c016123f31b86936a2046c42cc8380036fbcd6d008ba0c2801eb73884ce8b9.mp4_original"
               }

      assert reloaded_old.data["meta"] ==
               %{
                 "source_key" =>
                   "dd02ebe7ebfa42777f3265794c2274ad85447d40731908a78f6288f7dcdd930b.gif_original"
               }
    end
  end

  describe "don't update existing source_key" do
    test "on image", context do
      image_meta = Map.put(context.image_meta, "source_key", "old_source.jpg_original")
      object_data = Map.put(context.object_data, "meta", image_meta)
      {:ok, object} = Repo.insert(%Object{data: object_data})

      Attachments.run(["images"])

      reloaded = Repo.get(Object, object.id)

      assert reloaded.data["meta"] == image_meta
    end

    test "on video", context do
      video_meta = Map.put(context.video_meta, "source_key", "old_source.mp4_original")
      object_data = Map.put(context.object_data, "meta", video_meta)
      {:ok, object} = Repo.insert(%Object{data: object_data})

      Attachments.run(["videos"])

      reloaded = Repo.get(Object, object.id)

      assert reloaded.data["meta"] == video_meta
    end

    test "on gifs", context do
      gif_meta = %{
        "source_key" => "old_original.gif_original"
      }

      gif_data = Map.put(context.new_gif_data, "meta", gif_meta)

      {:ok, object} = Repo.insert(%Object{data: gif_data})

      Attachments.run(["gifs"])

      reloaded = Repo.get(Object, object.id)
      assert reloaded.data["meta"] == gif_meta
    end
  end
end
