# Pleroma: A lightweight social networking server
# Copyright © 2017-2019 Pleroma Authors <https://pleroma.social/>
# SPDX-License-Identifier: AGPL-3.0-only

defmodule Pleroma.Web.ActivityPub.MRF.CdnWarmingPolicyTest do
  use Pleroma.DataCase

  alias Pleroma.HTTP
  alias Pleroma.Tests.ObanHelpers
  alias Pleroma.Web.ActivityPub.MRF.CdnWarmingPolicy

  import Mock
  import ExUnit.CaptureLog

  @message %{
    "type" => "Create",
    "object" => %{
      "type" => "Note",
      "content" => "content",
      "attachment" => [
        %{
          "url" => [
            %{
              "href" => "https://pleroma-media.test.com/image.jpg",
              "preview_url" => "https://pleroma-media.test.com/image.jpg.preview.jpg"
            },
            %{
              "href" => "https://pleroma-media.test.com/image1.jpg"
            }
          ]
        }
      ]
    }
  }

  setup do: clear_config([CdnWarmingPolicy])

  test "it warming media URIs" do
    with_mock HTTP, get: fn _, _, _ -> {:ok, []} end do
      Pleroma.Config.put([CdnWarmingPolicy],
        cdn_trigger_endpoint: "http://master.cdn.cloud.com/edge/cache/warm",
        log_level: :debug
      )

      CdnWarmingPolicy.filter(@message)

      ObanHelpers.perform_all()
      # Performing jobs which has been just enqueued
      ObanHelpers.perform_all()
      headers = [{"HOST", "pleroma-media.test.com"}]

      [
        "http://master.cdn.cloud.com:80/edge/cache/warm/image.jpg",
        "http://master.cdn.cloud.com:80/edge/cache/warm/image.jpg.preview.jpg",
        "http://master.cdn.cloud.com:80/edge/cache/warm/image1.jpg"
      ]
      |> Enum.each(fn url ->
        assert called(HTTP.get(url, headers, :_))
      end)
    end
  end

  test "it does nothing when CDN is not set" do
    with_mock HTTP, get: fn _, _, _ -> {:ok, []} end do
      Pleroma.Config.put([Pleroma.Web.ActivityPub.MRF.CdnWarmingPolicy],
        cdn_trigger_endpoint: "",
        log_level: :debug
      )

      assert capture_log(fn ->
               CdnWarmingPolicy.filter(@message)
               ObanHelpers.perform_all()
               ObanHelpers.perform_all()
             end) =~
               "config :pleroma, Pleroma.Web.ActivityPub.MRF.CdnWarmingPolicy, :cdn_trigger_endpoint is not set"

      refute called(HTTP.get(:_, :_, :_))
    end
  end

  test "it does nothing when no attachments are present" do
    object =
      @message["object"]
      |> Map.delete("attachment")

    message =
      @message
      |> Map.put("object", object)

    with_mock HTTP, get: fn _, _, _ -> {:ok, []} end do
      CdnWarmingPolicy.filter(message)
      refute called(HTTP.get(:_, :_, :_))
    end
  end
end
