# Pleroma: A lightweight social networking server
# Copyright © 2017-2020 Pleroma Authors <https://pleroma.social/>
# SPDX-License-Identifier: AGPL-3.0-only

defmodule Pleroma.Web.Metadata.Providers.OpenGraphTest do
  use Pleroma.DataCase
  import Pleroma.Factory
  alias Pleroma.Web.Metadata.Providers.OpenGraph

  setup do: clear_config([Pleroma.Web.Metadata, :unfurl_nsfw])

  test "it renders all supported types of attachments and skips unknown types" do
    user = insert(:user)

    note =
      insert(:note, %{
        data: %{
          "actor" => user.ap_id,
          "tag" => [],
          "id" => "https://pleroma.gov/objects/whatever",
          "content" => "pleroma in a nutshell",
          "attachment" => [
            %{
              "url" => [
                %{"mediaType" => "image/png", "href" => "https://pleroma.gov/tenshi.png"}
              ]
            },
            %{
              "url" => [
                %{
                  "mediaType" => "application/octet-stream",
                  "href" => "https://pleroma.gov/fqa/badapple.sfc"
                }
              ]
            },
            %{
              "url" => [
                %{"mediaType" => "video/webm", "href" => "https://pleroma.gov/about/juche.webm"}
              ]
            },
            %{
              "url" => [
                %{
                  "mediaType" => "audio/basic",
                  "href" => "http://www.gnu.org/music/free-software-song.au"
                }
              ]
            }
          ]
        }
      })

    result = OpenGraph.build_tags(%{object: note, url: note.data["id"], user: user})

    assert Enum.all?(
             [
               {:meta, [property: "og:image", content: "https://pleroma.gov/tenshi.png"], []},
               {:meta,
                [property: "og:audio", content: "http://www.gnu.org/music/free-software-song.au"],
                []},
               {:meta, [property: "og:video", content: "https://pleroma.gov/about/juche.webm"],
                []}
             ],
             fn element -> element in result end
           )
  end

  test "it does not render attachments if post is nsfw" do
    Pleroma.Config.put([Pleroma.Web.Metadata, :unfurl_nsfw], false)
    user = insert(:user, avatar: %{"url" => [%{"href" => "https://pleroma.gov/tenshi.png"}]})

    note =
      insert(:note, %{
        data: %{
          "actor" => user.ap_id,
          "id" => "https://pleroma.gov/objects/whatever",
          "content" => "#cuteposting #nsfw #hambaga",
          "tag" => ["cuteposting", "nsfw", "hambaga"],
          "sensitive" => true,
          "attachment" => [
            %{
              "url" => [
                %{"mediaType" => "image/png", "href" => "https://misskey.microsoft/corndog.png"}
              ]
            }
          ]
        }
      })

    result = OpenGraph.build_tags(%{object: note, url: note.data["id"], user: user})

    assert {:meta, [property: "og:image", content: "https://pleroma.gov/tenshi.png"], []} in result

    refute {:meta, [property: "og:image", content: "https://misskey.microsoft/corndog.png"], []} in result
  end

  test "renders width and height tags for images" do
    user = insert(:user, avatar: %{"url" => [%{"href" => "https://pleroma.gov/tenshi.png"}]})

    note =
      insert(:note, %{
        data: %{
          "actor" => user.ap_id,
          "id" => "https://pleroma.gov/objects/whatever",
          "content" => "cofe",
          "tag" => [],
          "sensitive" => false,
          "attachment" => [
            %{
              "meta" => %{
                "small" => %{
                  "size" => "400x400",
                  "content_type" => "image/jpeg",
                  "method" => "resize",
                  "width" => 400,
                  "height" => 400,
                  "dest_key" =>
                    "9888b901b50ee63e675bba2431646a6be1364a301a0ddfdeae9c4dbc2fb584bb.jpg.preview.jpg"
                },
                "original" => %{
                  "size" => "1920x1080",
                  "content_type" => "image/jpeg",
                  "method" => "resize",
                  "width" => 1920,
                  "height" => 1080,
                  "dest_key" =>
                    "9888b901b50ee63e675bba2431646a6be1364a301a0ddfdeae9c4dbc2fb584bb.jpg"
                },
                "source_key" =>
                  "9888b901b50ee63e675bba2431646a6be1364a301a0ddfdeae9c4dbc2fb584bb.jpg_original"
              },
              "actor" => user.ap_id,
              "url" => [
                %{
                  "type" => "Link",
                  "href" =>
                    "http://pleroma.gov/media/9888b901b50ee63e675bba2431646a6be1364a301a0ddfdeae9c4dbc2fb584bb.jpg?name=IMGA0002.jpg",
                  "mediaType" => "image/jpeg"
                }
              ],
              "name" => "IMGA0002.jpg",
              "type" => "Document"
            }
          ]
        }
      })

    result = OpenGraph.build_tags(%{object: note, url: note.data["id"], user: user})

    assert {:meta, [property: "og:image:width", content: 1920], []} in result
    assert {:meta, [property: "og:image:height", content: 1080], []} in result
  end
end
