# Pleroma: A lightweight social networking server
# Copyright © 2017-2019 Pleroma Authors <https://pleroma.social/>
# SPDX-License-Identifier: AGPL-3.0-only

defmodule PleromaMfc.UserTest do
  alias Pleroma.ExNickname
  alias Pleroma.Repo
  alias Pleroma.User

  use Pleroma.DataCase
  use Oban.Testing, repo: Pleroma.Repo

  import Pleroma.Factory

  setup_all do
    Tesla.Mock.mock_global(fn env -> apply(HttpRequestMock, :request, [env]) end)
    :ok
  end

  describe "update_nickname/2" do
    setup do: clear_config([:instance, :federating], false)

    setup do: clear_config([:instance, :allow_local_nickname_changing], true)

    setup do
      user = insert(:user, local: true)
      {:ok, user: user}
    end

    test "no op if new nickname is the same as existing one", %{user: user} do
      assert {:ok, user} == User.update_nickname(user, user.nickname)
    end

    test "returns `{:error, :forbidden}` if instance is federating", %{user: user} do
      Pleroma.Config.put([:instance, :federating], true)

      assert {:error, :forbidden} == User.update_nickname(user, user.nickname <> "a")
    end

    test "returns `{:error, :forbidden}` if nickname changes are disabled", %{user: user} do
      Pleroma.Config.put([:instance, :allow_local_nickname_changing], false)

      assert {:error, :forbidden} == User.update_nickname(user, user.nickname <> "a")
    end

    test "returns `{:error, :unavailable}` if nickname is occupied by another user", %{user: user} do
      user2 = insert(:user)

      assert {:error, :unavailable} == User.update_nickname(user, user2.nickname)
    end

    test "returns `{:error, :unavailable}` if nickname was used by another user", %{user: user} do
      user2 = insert(:user)
      user2_ex_nickname = insert(:ex_nickname, user: user2)

      assert {:error, :unavailable} == User.update_nickname(user, user2_ex_nickname.nickname)
    end

    test "changes local user's `nickname` if nickname is valid and was never occupied", %{
      user: user
    } do
      old_nickname = user.nickname
      new_nickname = old_nickname <> "a"

      assert {:ok, %User{nickname: new_nickname} = updated_user} =
               User.update_nickname(user, new_nickname)

      assert updated_user == User.get_by_id(user.id)

      assert Repo.get_by(ExNickname, nickname: old_nickname)
      refute Repo.get_by(User, nickname: old_nickname)

      for alias <- [old_nickname, new_nickname] do
        assert updated_user == User.get_by_nickname(alias)
        assert updated_user == User.get_cached_by_nickname(alias)
        assert updated_user == User.get_cached_by_nickname_or_id(alias)
      end

      assert {:ok, nil} = Cachex.get(:user_cache, "nickname:#{old_nickname}")
      assert {:ok, updated_user} = Cachex.get(:user_cache, "nickname:#{new_nickname}")
    end

    test "changes local user's `nickname` if nickname was previously used by the same user", %{
      user: user
    } do
      ex_nickname_record = insert(:ex_nickname, user: user)
      reused_nickname = ex_nickname_record.nickname

      assert {:ok, updated_user} = User.update_nickname(user, reused_nickname)
      assert updated_user == User.get_by_id(user.id)
      assert updated_user.nickname == reused_nickname
      assert Repo.get_by(ExNickname, nickname: user.nickname)
      refute Repo.get_by(ExNickname, nickname: reused_nickname)
    end
  end
end
