# Pleroma: A lightweight social networking server
# Copyright © 2017-2018 Pleroma Authors <https://pleroma.social/>
# SPDX-License-Identifier: AGPL-3.0-only

defmodule PleromaMfc.Web.AdminAPI.AdminAPIControllerTest do
  use Pleroma.Web.ConnCase

  alias Pleroma.Mfc.Notify
  alias Pleroma.User

  import Pleroma.Factory

  setup do
    admin = insert(:user, is_admin: true)
    token = insert(:oauth_admin_token, user: admin)

    conn =
      build_conn()
      |> assign(:user, admin)
      |> assign(:token, token)

    {:ok, %{admin: admin, token: token, conn: conn}}
  end

  describe "/api/pleroma/admin/users/follow" do
    test "allows to force-follow another user", %{conn: conn} do
      user = insert(:user, mfc_id: "1")
      follower = insert(:user, mfc_id: "2")

      conn
      |> put_req_header("accept", "application/json")
      |> post("/api/pleroma/admin/users/follow", %{
        "follower" => follower.nickname,
        "followed" => user.nickname
      })

      user = User.get_cached_by_id(user.id)
      follower = User.get_cached_by_id(follower.id)

      assert User.following?(follower, user)

      url = Pleroma.Config.get([:mfc, :user_info_sync_endpoint])

      Tesla.Mock.mock(fn %{url: ^url, body: body} ->
        assert body =~ "mfc_id=1"
        assert body =~ "friends=0"
        assert body =~ "followers=1"
        assert body =~ "active=true"
        send(self(), :called_api)
        %Tesla.Env{status: 200}
      end)

      assert Notify.perform(%Oban.Job{args: %{"op" => "user_change", "user_id" => user.id}})
      assert_received(:called_api)

      Tesla.Mock.mock(fn %{url: ^url, body: body} ->
        assert body =~ "mfc_id=2"
        assert body =~ "friends=1"
        assert body =~ "followers=0"
        assert body =~ "active=true"
        send(self(), :called_api)
        %Tesla.Env{status: 200}
      end)

      assert Notify.perform(%Oban.Job{args: %{"op" => "user_change", "user_id" => follower.id}})
      assert_received(:called_api)
    end
  end

  describe "/api/pleroma/admin/users/unfollow" do
    test "allows to force-unfollow another user", %{conn: conn} do
      user = insert(:user, mfc_id: "1")
      follower = insert(:user, mfc_id: "2")

      User.follow(follower, user)

      conn
      |> put_req_header("accept", "application/json")
      |> post("/api/pleroma/admin/users/unfollow", %{
        "follower" => follower.nickname,
        "followed" => user.nickname
      })

      user = User.get_cached_by_id(user.id)
      follower = User.get_cached_by_id(follower.id)

      refute User.following?(follower, user)

      url = Pleroma.Config.get([:mfc, :user_info_sync_endpoint])

      Tesla.Mock.mock(fn %{url: ^url, body: body} ->
        assert body =~ "mfc_id=1"
        assert body =~ "friends=0"
        assert body =~ "followers=0"
        assert body =~ "active=true"
        send(self(), :called_api)
        %Tesla.Env{status: 200}
      end)

      assert Notify.perform(%Oban.Job{args: %{"op" => "user_change", "user_id" => user.id}})
      assert_received(:called_api)

      Tesla.Mock.mock(fn %{url: ^url, body: body} ->
        assert body =~ "mfc_id=2"
        assert body =~ "friends=0"
        assert body =~ "followers=0"
        assert body =~ "active=true"
        send(self(), :called_api)
        %Tesla.Env{status: 200}
      end)

      assert Notify.perform(%Oban.Job{args: %{"op" => "user_change", "user_id" => follower.id}})
      assert_received(:called_api)
    end
  end

  test "PATCH /api/pleroma/admin/users/deactivate", %{conn: conn} do
    user = insert(:user, deactivated: false, mfc_id: "1")

    conn =
      patch(
        conn,
        "/api/pleroma/admin/users/deactivate",
        %{nicknames: [user.nickname]}
      )

    user = User.get_cached_by_id(user.id)
    assert user.deactivated == true
    assert json_response(conn, 200)

    url = Pleroma.Config.get([:mfc, :user_info_sync_endpoint])

    Tesla.Mock.mock(fn %{url: ^url, body: body} ->
      assert body =~ "mfc_id=1"
      assert body =~ "friends=0"
      assert body =~ "followers=0"
      assert body =~ "active=false"
      send(self(), :called_api)
      %Tesla.Env{status: 200}
    end)

    assert Notify.perform(%Oban.Job{args: %{"op" => "user_change", "user_id" => user.id}})
    assert_received(:called_api)
  end

  test "PATCH /api/pleroma/admin/users/activate", %{conn: conn} do
    user = insert(:user, deactivated: true, mfc_id: "1")

    conn =
      patch(
        conn,
        "/api/pleroma/admin/users/activate",
        %{nicknames: [user.nickname]}
      )

    user = User.get_cached_by_id(user.id)
    assert user.deactivated == false
    assert json_response(conn, 200)

    url = Pleroma.Config.get([:mfc, :user_info_sync_endpoint])

    Tesla.Mock.mock(fn %{url: ^url, body: body} ->
      assert body =~ "mfc_id=1"
      assert body =~ "friends=0"
      assert body =~ "followers=0"
      assert body =~ "active=true"
      send(self(), :called_api)
      %Tesla.Env{status: 200}
    end)

    assert Notify.perform(%Oban.Job{args: %{"op" => "user_change", "user_id" => user.id}})
    assert_received(:called_api)
  end

  describe "PATCH /api/pleroma/admin/users/:current_nickname/update_nickname" do
    setup do: clear_config([:instance, :federating], false)

    setup do: clear_config([:instance, :allow_local_nickname_changing], true)

    test "local user's nickname can be changed", %{conn: conn} do
      user = insert(:user)
      new_nickname = user.nickname <> "a"

      conn =
        patch(conn, "/api/pleroma/admin/users/#{user.nickname}/update_nickname", %{
          nickname: new_nickname
        })

      assert json_response(conn, 200)

      user = refresh_record(user)
      assert new_nickname == user.nickname
    end

    test "local user's nickname can be changed by mfc id", %{conn: conn} do
      user = insert(:user, mfc_id: "1")
      new_nickname = user.nickname <> "a"

      conn =
        patch(conn, "/api/pleroma/admin/users/by_mfc_id/#{user.mfc_id}/update_nickname", %{
          nickname: new_nickname
        })

      assert json_response(conn, 200)

      user = refresh_record(user)
      assert new_nickname == user.nickname
    end

    test "renders error json if nickname validation failed, extended nickname format disabled", %{
      conn: conn
    } do
      clear_config([:instance, :extended_nickname_format], false)
      user = insert(:user)

      conn =
        patch(conn, "/api/pleroma/admin/users/#{user.nickname}/update_nickname", %{
          nickname: "_bad_nickname"
        })

      assert json_response(conn, 422) == %{"error" => "nickname has invalid format"}
    end

    test "renders error json if nickname validation failed, extended nickname format enabled", %{
      conn: conn
    } do
      clear_config([:instance, :extended_nickname_format], true)
      user = insert(:user)

      conn =
        patch(conn, "/api/pleroma/admin/users/#{user.nickname}/update_nickname", %{
          nickname: "^bad%nickname"
        })

      assert json_response(conn, 422) == %{"error" => "nickname has invalid format"}
    end
  end
end
